@@ -67,7 +67,7 @@ import qualified Distribution.Package as P
 import Distribution.Client.Init.Types
   ( InitFlags(..), PackageType(..), Category(..) )
 import Distribution.Client.Init.Licenses
-  ( bsd3, gplv2, gplv3, lgpl2, lgpl3 )
+  ( bsd3, gplv2, gplv3, lgpl2, lgpl3, apache20 )
 import Distribution.Client.Init.Heuristics
   ( guessPackageName, guessAuthorNameMail, SourceFileEntry(..), scanForModules, neededBuildPrograms )
 
@@ -178,7 +178,7 @@ getLicense flags = do
   return $ flags { license = maybeToFlag lic }
   where
     listedLicenses =
-      knownLicenses \\ [GPL Nothing, LGPL Nothing, OtherLicense]
+      knownLicenses \\ [GPL Nothing, LGPL Nothing, Apache Nothing, OtherLicense]
 
 -- | The author's name and email. Prompt, or try to guess from an existing
 --   darcs repo.
@@ -495,6 +495,9 @@ writeLicense flags = do
           Flag (LGPL (Just (Version {versionBranch = [3]})))
             -> Just lgpl3
 
+          Flag (Apache (Just (Version {versionBranch = [2, 0]})))
+            -> Just apache20
+
           _ -> Nothing
 
   case licenseFile of
