@@ -15,11 +15,11 @@ module Distribution.Client.List (
 
 import Distribution.Package
          ( PackageName(..), Package(..), packageName, packageVersion
-         , Dependency(..), thisPackageVersion, depends, simplifyDependency )
+         , Dependency(..), simplifyDependency )
 import Distribution.ModuleName (ModuleName)
 import Distribution.License (License)
 import qualified Distribution.InstalledPackageInfo as Installed
-import qualified Distribution.PackageDescription   as Available
+import qualified Distribution.PackageDescription   as Source
 import Distribution.PackageDescription
          ( Flag(..), FlagName(..) )
 import Distribution.PackageDescription.Configuration
@@ -31,6 +31,7 @@ import Distribution.Simple.Program (ProgramConfiguration)
 import Distribution.Simple.Utils
         ( equating, comparing, die, notice )
 import Distribution.Simple.Setup (fromFlag)
+import qualified Distribution.Simple.PackageIndex as InstalledPackageIndex
 import qualified Distribution.Client.PackageIndex as PackageIndex
 import Distribution.Version
          ( Version(..), VersionRange, withinRange, anyVersion
@@ -40,10 +41,9 @@ import Distribution.Text
          ( Text(disp), display )
 
 import Distribution.Client.Types
-         ( AvailablePackage(..), Repo, AvailablePackageDb(..)
-         , InstalledPackage(..) )
+         ( SourcePackage(..), Repo, SourcePackageDb(..) )
 import Distribution.Client.Dependency.Types
-         ( PackageConstraint(..) )
+         ( PackageConstraint(..), ExtDependency(..) )
 import Distribution.Client.Targets
          ( UserTarget, resolveUserTargets, PackageSpecifier(..) )
 import Distribution.Client.Setup
@@ -51,7 +51,7 @@ import Distribution.Client.Setup
 import Distribution.Client.Utils
          ( mergeBy, MergeResult(..) )
 import Distribution.Client.IndexUtils as IndexUtils
-         ( getAvailablePackages, getInstalledPackages )
+         ( getSourcePackages, getInstalledPackages )
 import Distribution.Client.FetchUtils
          ( isFetched )
 
@@ -65,7 +65,7 @@ import Control.Monad
          ( MonadPlus(mplus), join )
 import Control.Exception
          ( assert )
-import Text.PrettyPrint.HughesPJ as Disp
+import Text.PrettyPrint as Disp
 import System.Directory
          ( doesDirectoryExist )
 
@@ -81,29 +81,29 @@ list :: Verbosity
      -> IO ()
 list verbosity packageDBs repos comp conf listFlags pats = do
 
-    installed   <- getInstalledPackages verbosity comp packageDBs conf
-    availableDb <- getAvailablePackages verbosity repos
-    let available  = packageIndex availableDb
+    installedPkgIndex <- getInstalledPackages verbosity comp packageDBs conf
+    sourcePkgDb       <- getSourcePackages    verbosity repos
+    let sourcePkgIndex = packageIndex sourcePkgDb
         prefs name = fromMaybe anyVersion
-                       (Map.lookup name (packagePreferences availableDb))
+                       (Map.lookup name (packagePreferences sourcePkgDb))
 
-        pkgsInfo :: [(PackageName, [InstalledPackage], [AvailablePackage])]
+        pkgsInfo :: [(PackageName, [Installed.InstalledPackageInfo], [SourcePackage])]
         pkgsInfo
             -- gather info for all packages
-          | null pats = mergePackages (PackageIndex.allPackages installed)
-                                      (PackageIndex.allPackages available)
+          | null pats = mergePackages (InstalledPackageIndex.allPackages installedPkgIndex)
+                                      (         PackageIndex.allPackages sourcePkgIndex)
 
             -- gather info for packages matching search term
-          | otherwise = mergePackages (matchingPackages installed)
-                                      (matchingPackages available)
+          | otherwise = mergePackages (matchingPackages InstalledPackageIndex.searchByNameSubstring installedPkgIndex)
+                                      (matchingPackages (\ idx n -> concatMap snd (PackageIndex.searchByNameSubstring idx n)) sourcePkgIndex)
 
         matches :: [PackageDisplayInfo]
         matches = [ mergePackageInfo pref
-                      installedPkgs availablePkgs selectedPkg False
-                  | (pkgname, installedPkgs, availablePkgs) <- pkgsInfo
+                      installedPkgs sourcePkgs selectedPkg False
+                  | (pkgname, installedPkgs, sourcePkgs) <- pkgsInfo
                   , not onlyInstalled || not (null installedPkgs)
                   , let pref        = prefs pkgname
-                        selectedPkg = latestWithPref pref availablePkgs ]
+                        selectedPkg = latestWithPref pref sourcePkgs ]
 
     if simpleOutput
       then putStr $ unlines
@@ -112,10 +112,10 @@ list verbosity packageDBs repos comp conf listFlags pats = do
              , version <- if onlyInstalled
                             then              installedVersions pkg
                             else nub . sort $ installedVersions pkg
-                                           ++ availableVersions pkg ]
+                                           ++ sourceVersions    pkg ]
              -- Note: this only works because for 'list', one cannot currently
              -- specify any version constraints, so listing all installed
-             -- and available ones works.
+             -- and source ones works.
       else
         if null matches
             then notice verbosity "No matches found."
@@ -124,11 +124,10 @@ list verbosity packageDBs repos comp conf listFlags pats = do
     onlyInstalled = fromFlag (listInstalled listFlags)
     simpleOutput  = fromFlag (listSimpleOutput listFlags)
 
-    matchingPackages index =
+    matchingPackages search index =
       [ pkg
       | pat <- pats
-      , (_, pkgs) <- PackageIndex.searchByNameSubstring index pat
-      , pkg <- pkgs ]
+      , pkg <- search index pat ]
 
 info :: Verbosity
      -> PackageDBStack
@@ -142,66 +141,73 @@ info :: Verbosity
 info verbosity packageDBs repos comp conf
      globalFlags _listFlags userTargets = do
 
-    installed     <- getInstalledPackages verbosity comp packageDBs conf
-    availableDb   <- getAvailablePackages verbosity repos
-    let available  = packageIndex availableDb
+    installedPkgIndex <- getInstalledPackages verbosity comp packageDBs conf
+    sourcePkgDb   <- getSourcePackages    verbosity repos
+    let sourcePkgIndex = packageIndex sourcePkgDb
         prefs name = fromMaybe anyVersion
-                       (Map.lookup name (packagePreferences availableDb))
+                       (Map.lookup name (packagePreferences sourcePkgDb))
 
         -- Users may specify names of packages that are only installed, not
         -- just available source packages, so we must resolve targets using
-        -- the combination of installed and available packages.
-    let available' = PackageIndex.fromList
-                   $ map packageId (PackageIndex.allPackages installed)
-                  ++ map packageId (PackageIndex.allPackages available)
+        -- the combination of installed and source packages.
+    let sourcePkgs' = PackageIndex.fromList
+                    $ map packageId (InstalledPackageIndex.allPackages installedPkgIndex)
+                   ++ map packageId (         PackageIndex.allPackages sourcePkgIndex)
     pkgSpecifiers <- resolveUserTargets verbosity
-                       globalFlags available' userTargets
+                       (fromFlag $ globalWorldFile globalFlags)
+                       sourcePkgs' userTargets
 
     pkgsinfo      <- sequence
                        [ do pkginfo <- either die return $
                                          gatherPkgInfo prefs
-                                           installed available pkgSpecifier
+                                           installedPkgIndex sourcePkgIndex
+                                           pkgSpecifier
                             updateFileSystemPackageDetails pkginfo
                        | pkgSpecifier <- pkgSpecifiers ]
 
     putStr $ unlines (map showPackageDetailedInfo pkgsinfo)
 
   where
-    gatherPkgInfo prefs installed available (NamedPackage name constraints)
-      | null (selectedInstalledPkgs) && null (selectedAvailablePkgs)
+    gatherPkgInfo :: (PackageName -> VersionRange) ->
+                     InstalledPackageIndex.PackageIndex ->
+                     PackageIndex.PackageIndex SourcePackage ->
+                     PackageSpecifier SourcePackage ->
+                     Either String PackageDisplayInfo
+    gatherPkgInfo prefs installedPkgIndex sourcePkgIndex (NamedPackage name constraints)
+      | null (selectedInstalledPkgs) && null (selectedSourcePkgs)
       = Left $ "There is no available version of " ++ display name
             ++ " that satisfies "
             ++ display (simplifyVersionRange verConstraint)
 
       | otherwise
       = Right $ mergePackageInfo pref installedPkgs
-                                 availablePkgs  selectedAvailablePkg
+                                 sourcePkgs  selectedSourcePkg'
                                  showPkgVersion
       where
         pref           = prefs name
-        installedPkgs  = PackageIndex.lookupPackageName installed name
-        availablePkgs  = PackageIndex.lookupPackageName available name
+        installedPkgs  = concatMap snd (InstalledPackageIndex.lookupPackageName installedPkgIndex name)
+        sourcePkgs     =                         PackageIndex.lookupPackageName sourcePkgIndex name
 
-        selectedInstalledPkgs = PackageIndex.lookupDependency installed
+        selectedInstalledPkgs = InstalledPackageIndex.lookupDependency installedPkgIndex
                                     (Dependency name verConstraint)
-        selectedAvailablePkgs = PackageIndex.lookupDependency available
+        selectedSourcePkgs    =          PackageIndex.lookupDependency sourcePkgIndex
                                     (Dependency name verConstraint)
-        selectedAvailablePkg  = latestWithPref pref selectedAvailablePkgs
+        selectedSourcePkg'    = latestWithPref pref selectedSourcePkgs
 
                          -- display a specific package version if the user
                          -- supplied a non-trivial version constraint
         showPkgVersion = not (null verConstraints)
         verConstraint  = foldr intersectVersionRanges anyVersion verConstraints
-        verConstraints = [ vr | PackageVersionConstraint _ vr <- constraints ]
+        verConstraints = [ vr | PackageConstraintVersion _ vr <- constraints ]
 
-    gatherPkgInfo prefs installed available (SpecificSourcePackage pkg) =
-        Right $ mergePackageInfo pref installedPkgs availablePkgs
+    gatherPkgInfo prefs installedPkgIndex sourcePkgIndex (SpecificSourcePackage pkg) =
+        Right $ mergePackageInfo pref installedPkgs sourcePkgs
                                  selectedPkg True
       where
         name          = packageName pkg
         pref          = prefs name
-        installedPkgs = PackageIndex.lookupPackageName installed name
-        availablePkgs = PackageIndex.lookupPackageName available name
+        installedPkgs = concatMap snd (InstalledPackageIndex.lookupPackageName installedPkgIndex name)
+        sourcePkgs    =                         PackageIndex.lookupPackageName sourcePkgIndex name
         selectedPkg   = Just pkg
 
 
@@ -211,9 +217,9 @@ info verbosity packageDBs repos comp conf
 data PackageDisplayInfo = PackageDisplayInfo {
     pkgName           :: PackageName,
     selectedVersion   :: Maybe Version,
-    selectedAvailable :: Maybe AvailablePackage,
+    selectedSourcePkg :: Maybe SourcePackage,
     installedVersions :: [Version],
-    availableVersions :: [Version],
+    sourceVersions    :: [Version],
     preferredVersions :: VersionRange,
     homepage          :: String,
     bugReports        :: String,
@@ -224,7 +230,7 @@ data PackageDisplayInfo = PackageDisplayInfo {
     license           :: License,
     author            :: String,
     maintainer        :: String,
-    dependencies      :: [Dependency],
+    dependencies      :: [ExtDependency],
     flags             :: [Flag],
     hasLib            :: Bool,
     hasExe            :: Bool,
@@ -242,7 +248,7 @@ showPackageSummaryInfo pkginfo =
      (nest 4 $ vcat [
        maybeShow (synopsis pkginfo) "Synopsis:" reflowParagraphs
      , text "Default available version:" <+>
-       case selectedAvailable pkginfo of
+       case selectedSourcePkg pkginfo of
          Nothing  -> text "[ Not available from any configured repository ]"
          Just pkg -> disp (packageVersion pkg)
      , text "Installed versions:" <+>
@@ -269,7 +275,7 @@ showPackageDetailedInfo pkginfo =
    $+$
    (nest 4 $ vcat [
      entry "Synopsis"      synopsis     hideIfNull     reflowParagraphs
-   , entry "Versions available" availableVersions
+   , entry "Versions available" sourceVersions
            (altText null "[ Not available from server ]")
            (dispTopVersions 9 (preferredVersions pkginfo))
    , entry "Versions installed" installedVersions
@@ -351,56 +357,55 @@ reflowLines = vcat . map text . lines
 -- package name.
 --
 mergePackageInfo :: VersionRange
-                 -> [InstalledPackage]
-                 -> [AvailablePackage]
-                 -> Maybe AvailablePackage
+                 -> [Installed.InstalledPackageInfo]
+                 -> [SourcePackage]
+                 -> Maybe SourcePackage
                  -> Bool
                  -> PackageDisplayInfo
-mergePackageInfo versionPref installedPkgs availablePkgs selectedPkg showVer =
-  assert (length installedPkgs + length availablePkgs > 0) $
+mergePackageInfo versionPref installedPkgs sourcePkgs selectedPkg showVer =
+  assert (length installedPkgs + length sourcePkgs > 0) $
   PackageDisplayInfo {
-    pkgName           = combine packageName available
+    pkgName           = combine packageName source
                                 packageName installed,
     selectedVersion   = if showVer then fmap packageVersion selectedPkg
                                    else Nothing,
-    selectedAvailable = availableSelected,
+    selectedSourcePkg = sourceSelected,
     installedVersions = map packageVersion installedPkgs,
-    availableVersions = map packageVersion availablePkgs,
+    sourceVersions    = map packageVersion sourcePkgs,
     preferredVersions = versionPref,
 
-    license      = combine Available.license    available
+    license      = combine Source.license       source
                            Installed.license    installed,
-    maintainer   = combine Available.maintainer available
+    maintainer   = combine Source.maintainer    source
                            Installed.maintainer installed,
-    author       = combine Available.author     available
+    author       = combine Source.author        source
                            Installed.author     installed,
-    homepage     = combine Available.homepage   available
+    homepage     = combine Source.homepage      source
                            Installed.homepage   installed,
-    bugReports   = maybe "" Available.bugReports available,
+    bugReports   = maybe "" Source.bugReports source,
     sourceRepo   = fromMaybe "" . join
-                 . fmap (uncons Nothing Available.repoLocation
-                       . sortBy (comparing Available.repoKind)
-                       . Available.sourceRepos)
-                 $ available,
+                 . fmap (uncons Nothing Source.repoLocation
+                       . sortBy (comparing Source.repoKind)
+                       . Source.sourceRepos)
+                 $ source,
                     --TODO: installed package info is missing synopsis
-    synopsis     = maybe "" Available.synopsis   available,
-    description  = combine Available.description available
+    synopsis     = maybe "" Source.synopsis      source,
+    description  = combine Source.description    source
                            Installed.description installed,
-    category     = combine Available.category    available
+    category     = combine Source.category       source
                            Installed.category    installed,
-    flags        = maybe [] Available.genPackageFlags availableGeneric,
+    flags        = maybe [] Source.genPackageFlags sourceGeneric,
     hasLib       = isJust installed
                 || fromMaybe False
-                   (fmap (isJust . Available.condLibrary) availableGeneric),
+                   (fmap (isJust . Source.condLibrary) sourceGeneric),
     hasExe       = fromMaybe False
-                   (fmap (not . null . Available.condExecutables) availableGeneric),
-    executables  = map fst (maybe [] Available.condExecutables availableGeneric),
+                   (fmap (not . null . Source.condExecutables) sourceGeneric),
+    executables  = map fst (maybe [] Source.condExecutables sourceGeneric),
     modules      = combine Installed.exposedModules installed
-                           (maybe [] Available.exposedModules
-                                   . Available.library) available,
-    dependencies = map simplifyDependency
-                 $ combine Available.buildDepends available
-                           (map thisPackageVersion . depends) installed',
+                           (maybe [] Source.exposedModules
+                                   . Source.library) source,
+    dependencies = combine (map (SourceDependency . simplifyDependency) . Source.buildDepends) source
+                           (map InstalledDependency . Installed.depends) installed,
     haddockHtml  = fromMaybe "" . join
                  . fmap (listToMaybe . Installed.haddockHTMLs)
                  $ installed,
@@ -408,14 +413,14 @@ mergePackageInfo versionPref installedPkgs availablePkgs selectedPkg showVer =
   }
   where
     combine f x g y  = fromJust (fmap f x `mplus` fmap g y)
-    installed'       = latestWithPref versionPref installedPkgs
-    installed        = fmap (\(InstalledPackage p _) -> p) installed'
+    installed :: Maybe Installed.InstalledPackageInfo
+    installed = latestWithPref versionPref installedPkgs
 
-    availableSelected
+    sourceSelected
       | isJust selectedPkg = selectedPkg
-      | otherwise          = latestWithPref versionPref availablePkgs
-    availableGeneric = fmap packageDescription availableSelected
-    available        = fmap flattenPackageDescription availableGeneric
+      | otherwise          = latestWithPref versionPref sourcePkgs
+    sourceGeneric = fmap packageDescription sourceSelected
+    source        = fmap flattenPackageDescription sourceGeneric
 
     uncons :: b -> (a -> b) -> [a] -> b
     uncons z _ []    = z
@@ -429,7 +434,7 @@ mergePackageInfo versionPref installedPkgs availablePkgs selectedPkg showVer =
 updateFileSystemPackageDetails :: PackageDisplayInfo -> IO PackageDisplayInfo
 updateFileSystemPackageDetails pkginfo = do
   fetched   <- maybe (return False) (isFetched . packageSource)
-                     (selectedAvailable pkginfo)
+                     (selectedSourcePkg pkginfo)
   docsExist <- doesDirectoryExist (haddockHtml pkginfo)
   return pkginfo {
     haveTarball = fetched,
@@ -444,20 +449,20 @@ latestWithPref pref pkgs = Just (maximumBy (comparing prefThenVersion) pkgs)
                            in (withinRange ver pref, ver)
 
 
--- | Rearrange installed and available packages into groups referring to the
+-- | Rearrange installed and source packages into groups referring to the
 -- same package by name. In the result pairs, the lists are guaranteed to not
 -- both be empty.
 --
-mergePackages :: [InstalledPackage]
-              -> [AvailablePackage]
+mergePackages :: [Installed.InstalledPackageInfo]
+              -> [SourcePackage]
               -> [( PackageName
-                  , [InstalledPackage]
-                  , [AvailablePackage] )]
-mergePackages installed available =
+                  , [Installed.InstalledPackageInfo]
+                  , [SourcePackage] )]
+mergePackages installedPkgs sourcePkgs =
     map collect
   $ mergeBy (\i a -> fst i `compare` fst a)
-            (groupOn packageName installed)
-            (groupOn packageName available)
+            (groupOn packageName installedPkgs)
+            (groupOn packageName sourcePkgs)
   where
     collect (OnlyInLeft  (name,is)         ) = (name, is, [])
     collect (    InBoth  (_,is)   (name,as)) = (name, is, as)
