@@ -18,7 +18,10 @@ import Distribution.Package
 import Distribution.InstalledPackageInfo
          ( InstalledPackageInfo )
 import Distribution.PackageDescription
-         ( GenericPackageDescription, FlagAssignment )
+         ( Benchmark(..), GenericPackageDescription(..), FlagAssignment
+         , TestSuite(..) )
+import Distribution.PackageDescription.Configuration
+         ( mapTreeData )
 import Distribution.Client.PackageIndex
          ( PackageIndex )
 import Distribution.Version
@@ -26,6 +29,7 @@ import Distribution.Version
 
 import Data.Map (Map)
 import Network.URI (URI)
+import Data.ByteString.Lazy (ByteString)
 import Distribution.Compat.Exception
          ( SomeException )
 
@@ -34,8 +38,8 @@ newtype Password = Password { unPassword :: String }
 
 -- | This is the information we get from a @00-index.tar.gz@ hackage index.
 --
-data AvailablePackageDb = AvailablePackageDb {
-  packageIndex       :: PackageIndex AvailablePackage,
+data SourcePackageDb = SourcePackageDb {
+  packageIndex       :: PackageIndex SourcePackage,
   packagePreferences :: Map PackageName VersionRange
 }
 
@@ -72,31 +76,56 @@ instance PackageFixedDeps InstalledPackage where
 -- final configure process will be independent of the environment.
 --
 data ConfiguredPackage = ConfiguredPackage
-       AvailablePackage    -- package info, including repo
+       SourcePackage       -- package info, including repo
        FlagAssignment      -- complete flag assignment for the package
+       [OptionalStanza]    -- list of enabled optional stanzas for the package
        [PackageId]         -- set of exact dependencies. These must be
                            -- consistent with the 'buildDepends' in the
-                           -- 'PackageDescrption' that you'd get by applying
-                           -- the flag assignment.
+                           -- 'PackageDescription' that you'd get by applying
+                           -- the flag assignment and optional stanzas.
   deriving Show
 
 instance Package ConfiguredPackage where
-  packageId (ConfiguredPackage pkg _ _) = packageId pkg
+  packageId (ConfiguredPackage pkg _ _ _) = packageId pkg
 
 instance PackageFixedDeps ConfiguredPackage where
-  depends (ConfiguredPackage _ _ deps) = deps
+  depends (ConfiguredPackage _ _ _ deps) = deps
 
 
--- | We re-use @GenericPackageDescription@ and use the @package-url@
--- field to store the tarball URI.
-data AvailablePackage = AvailablePackage {
-    packageInfoId      :: PackageId,
-    packageDescription :: GenericPackageDescription,
-    packageSource      :: PackageLocation (Maybe FilePath)
+-- | A package description along with the location of the package sources.
+--
+data SourcePackage = SourcePackage {
+    packageInfoId        :: PackageId,
+    packageDescription   :: GenericPackageDescription,
+    packageSource        :: PackageLocation (Maybe FilePath),
+    packageDescrOverride :: PackageDescriptionOverride
   }
   deriving Show
 
-instance Package AvailablePackage where packageId = packageInfoId
+-- | We sometimes need to override the .cabal file in the tarball with
+-- the newer one from the package index.
+type PackageDescriptionOverride = Maybe ByteString
+
+instance Package SourcePackage where packageId = packageInfoId
+
+data OptionalStanza
+    = TestStanzas
+    | BenchStanzas
+  deriving (Eq, Ord, Show)
+
+enableStanzas
+    :: [OptionalStanza]
+    -> GenericPackageDescription
+    -> GenericPackageDescription
+enableStanzas stanzas gpkg = gpkg
+    { condBenchmarks = flagBenchmarks $ condBenchmarks gpkg
+    , condTestSuites = flagTests $ condTestSuites gpkg
+    }
+  where
+    enableTest t = t { testEnabled = TestStanzas `elem` stanzas }
+    enableBenchmark bm = bm { benchmarkEnabled = BenchStanzas `elem` stanzas }
+    flagBenchmarks = map (\(n, bm) -> (n, mapTreeData enableBenchmark bm))
+    flagTests = map (\(n, t) -> (n, mapTreeData enableTest t))
 
 -- ------------------------------------------------------------
 -- * Package locations and repositories
@@ -156,8 +185,9 @@ data BuildFailure = DependentFailed PackageId
                   | UnpackFailed    SomeException
                   | ConfigureFailed SomeException
                   | BuildFailed     SomeException
+                  | TestsFailed     SomeException
                   | InstallFailed   SomeException
 data BuildSuccess = BuildOk         DocsResult TestsResult
 
 data DocsResult  = DocsNotTried  | DocsFailed  | DocsOk
-data TestsResult = TestsNotTried | TestsFailed | TestsOk
+data TestsResult = TestsNotTried | TestsOk
