@@ -19,11 +19,11 @@ module Distribution.Client.Unpack (
   ) where
 
 import Distribution.Package
-         ( PackageId, packageId )
+         ( PackageId, packageId, packageName )
 import Distribution.Simple.Setup
-         ( fromFlagOrDefault )
+         ( fromFlag, fromFlagOrDefault )
 import Distribution.Simple.Utils
-         ( notice, die )
+         ( notice, die, info, writeFileAtomic )
 import Distribution.Verbosity
          ( Verbosity )
 import Distribution.Text(display)
@@ -36,7 +36,7 @@ import Distribution.Client.Dependency
 import Distribution.Client.FetchUtils
 import qualified Distribution.Client.Tar as Tar (extractTarGzFile)
 import Distribution.Client.IndexUtils as IndexUtils
-        ( getAvailablePackages )
+        ( getSourcePackages )
 
 import System.Directory
          ( createDirectoryIfMissing, doesDirectoryExist, doesFileExist )
@@ -45,8 +45,9 @@ import Control.Monad
 import Data.Monoid
          ( mempty )
 import System.FilePath
-         ( (</>), addTrailingPathSeparator )
-
+         ( (</>), (<.>), addTrailingPathSeparator )
+import qualified Data.ByteString.Lazy.Char8 as BS
+         ( unpack )
 
 unpack :: Verbosity
        -> [Repo]
@@ -60,14 +61,16 @@ unpack verbosity _ _ _ [] =
 unpack verbosity repos globalFlags unpackFlags userTargets = do
   mapM_ checkTarget userTargets
 
-  availableDb   <- getAvailablePackages verbosity repos
+  sourcePkgDb   <- getSourcePackages verbosity repos
 
   pkgSpecifiers <- resolveUserTargets verbosity
-                     globalFlags (packageIndex availableDb) userTargets
+                     (fromFlag $ globalWorldFile globalFlags)
+                     (packageIndex sourcePkgDb)
+                     userTargets
 
   pkgs <- either (die . unlines . map show) return $
             resolveWithoutDependencies
-              (resolverParams availableDb pkgSpecifiers)
+              (resolverParams sourcePkgDb pkgSpecifiers)
 
   unless (null prefix) $
          createDirectoryIfMissing True prefix
@@ -75,26 +78,29 @@ unpack verbosity repos globalFlags unpackFlags userTargets = do
   flip mapM_ pkgs $ \pkg -> do
     location <- fetchPackage verbosity (packageSource pkg)
     let pkgid = packageId pkg
+        descOverride | usePristine = Nothing
+                     | otherwise   = packageDescrOverride pkg
     case location of
       LocalTarballPackage tarballPath ->
-        unpackPackage verbosity prefix pkgid tarballPath
+        unpackPackage verbosity prefix pkgid descOverride tarballPath
 
       RemoteTarballPackage _tarballURL tarballPath ->
-        unpackPackage verbosity prefix pkgid tarballPath
+        unpackPackage verbosity prefix pkgid descOverride tarballPath
 
       RepoTarballPackage _repo _pkgid tarballPath ->
-        unpackPackage verbosity prefix pkgid tarballPath
+        unpackPackage verbosity prefix pkgid descOverride tarballPath
 
       LocalUnpackedPackage _ ->
         error "Distribution.Client.Unpack.unpack: the impossible happened."
 
   where
-    resolverParams availableDb pkgSpecifiers =
+    resolverParams sourcePkgDb pkgSpecifiers =
         --TODO: add commandline constraint and preference args for unpack
 
-        standardInstallPolicy mempty availableDb pkgSpecifiers
+        standardInstallPolicy mempty sourcePkgDb pkgSpecifiers
 
     prefix = fromFlagOrDefault "" (unpackDestDir unpackFlags)
+    usePristine = fromFlagOrDefault False (unpackPristine unpackFlags)
 
 checkTarget :: UserTarget -> IO ()
 checkTarget target = case target of
@@ -106,8 +112,10 @@ checkTarget target = case target of
         "The 'unpack' command is for tarball packages. "
      ++ "The target '" ++ t ++ "' is not a tarball."
 
-unpackPackage :: Verbosity -> FilePath -> PackageId -> FilePath -> IO ()
-unpackPackage verbosity prefix pkgid pkgPath = do
+unpackPackage :: Verbosity -> FilePath -> PackageId
+              -> PackageDescriptionOverride
+              -> FilePath  -> IO ()
+unpackPackage verbosity prefix pkgid descOverride pkgPath = do
     let pkgdirname = display pkgid
         pkgdir     = prefix </> pkgdirname
         pkgdir'    = addTrailingPathSeparator pkgdir
@@ -119,3 +127,12 @@ unpackPackage verbosity prefix pkgid pkgPath = do
      "A file \"" ++ pkgdir ++ "\" is in the way, not unpacking."
     notice verbosity $ "Unpacking to " ++ pkgdir'
     Tar.extractTarGzFile prefix pkgdirname pkgPath
+
+    case descOverride of
+      Nothing     -> return ()
+      Just pkgtxt -> do
+        let descFilePath = pkgdir </> display (packageName pkgid) <.> "cabal"
+        info verbosity $
+          "Updating " ++ descFilePath
+                      ++ " with the latest revision from the index."
+        writeFileAtomic descFilePath (BS.unpack pkgtxt)
